from dataclasses import dataclass, field

from mashumaro.mixins.orjson import DataClassORJSONMixin

from pymammotion.utility.device_type import DeviceType


@dataclass
class OperationSettings(DataClassORJSONMixin):
    """Operation settings for a device."""

    is_mow: bool = True
    is_dump: bool = True
    is_edge: bool = False
    collect_grass_frequency: int = 10
    job_mode: int = 4  # taskMode
    job_version: int = 0
    job_id: int = 0
    speed: float = 0.3
    ultra_wave: int = 2  # touch no touch etc
    channel_mode: int = 0  # grid or border first
    channel_width: int = 25
    rain_tactics: int = 0
    blade_height: int = 0
    path_order: str = ""
    toward: int = 0  # is just angle
    toward_included_angle: int = 90
    toward_mode: int = 0  # angle type relative etc
    border_mode: int = 0
    obstacle_laps: int = 1
    mowing_laps: int = 1  # border laps
    start_progress: int = 0
    areas: list[int] = field(default_factory=list)


def create_path_order(operation_mode: OperationSettings, device_name: str) -> str:
    # TODO add scheduling logic from getReserved() WorkSettingViewModel.java
    i2 = 0
    bArr = bytearray(8)
    bArr[0] = operation_mode.border_mode
    bArr[1] = operation_mode.obstacle_laps
    bArr[3] = int(operation_mode.start_progress)
    bArr[2] = 0
    if not DeviceType.is_luba1(device_name):
        bArr[4] = 0
        if DeviceType.is_yuka(device_name):
            bArr[5] = calculate_yuka_mode(operation_mode)
        else:
            bArr[5] = 8 if DeviceType.is_luba_2(device_name) else 0

        bArr[6] = int(operation_mode.collect_grass_frequency) if operation_mode.is_dump else 10
    if DeviceType.is_luba1(device_name):
        bArr[4] = operation_mode.toward_mode
    return bArr.decode()


def calculate_yuka_mode(operation_mode: OperationSettings) -> int:
    if operation_mode.is_mow and operation_mode.is_dump and operation_mode.is_edge:
        return 14
    if operation_mode.is_mow and operation_mode.is_dump and not operation_mode.is_edge:
        return 12
    if operation_mode.is_mow and not operation_mode.is_dump and operation_mode.is_edge:
        return 10
    if operation_mode.is_mow and not operation_mode.is_dump and not operation_mode.is_edge:
        return 8
    if not operation_mode.is_mow and operation_mode.is_dump and operation_mode.is_edge:
        return 6
    if not operation_mode.is_mow and not operation_mode.is_dump and operation_mode.is_edge:
        return 2
    if not operation_mode.is_mow and operation_mode.is_dump and not operation_mode.is_edge:
        return 4
    return 0



"""Module for interacting with Aliyun Cloud IoT Gateway."""

import base64
import hashlib
import hmac
import itertools
import json
from logging import getLogger
import random
import string
import time
import uuid

from aiohttp import ClientSession
from alibabacloud_iot_api_gateway.client import Client
from alibabacloud_iot_api_gateway.models import CommonParams, Config, IoTApiRequest
from alibabacloud_tea_util.client import Client as UtilClient
from alibabacloud_tea_util.models import RuntimeOptions

from pymammotion.aliyun.model.aep_response import AepResponse
from pymammotion.aliyun.model.connect_response import ConnectResponse
from pymammotion.aliyun.model.dev_by_account_response import ListingDevByAccountResponse
from pymammotion.aliyun.model.login_by_oauth_response import LoginByOAuthResponse
from pymammotion.aliyun.model.regions_response import RegionResponse
from pymammotion.aliyun.model.session_by_authcode_response import SessionByAuthCodeResponse
from pymammotion.const import ALIYUN_DOMAIN, APP_KEY, APP_SECRET, APP_VERSION
from pymammotion.http.http import MammotionHTTP
from pymammotion.utility.datatype_converter import DatatypeConverter

logger = getLogger(__name__)

MOVE_HEADERS = (
    "x-ca-signature",
    "x-ca-signature-headers",
    "accept",
    "content-md5",
    "content-type",
    "date",
    "host",
    "token",
    "user-agent",
)


class SetupException(Exception):
    """Raise when mqtt expires token or token is invalid."""


class AuthRefreshException(Exception):
    """Raise exception when library cannot refresh token."""


class DeviceOfflineException(Exception):
    """Raise exception when device is offline."""


class GatewayTimeoutException(Exception):
    """Raise exception when the gateway times out."""


class LoginException(Exception):
    """Raise exception when library cannot log in."""


class CheckSessionException(Exception):
    """Raise exception when checking session results in a failure."""


class CloudIOTGateway:
    """Class for interacting with Aliyun Cloud IoT Gateway."""

    _client_id = ""
    _device_sn = ""
    _utdid = ""

    converter = DatatypeConverter()

    def __init__(
        self,
        connect_response: ConnectResponse | None = None,
        login_by_oauth_response: LoginByOAuthResponse | None = None,
        aep_response: AepResponse | None = None,
        session_by_authcode_response: SessionByAuthCodeResponse | None = None,
        region_response: RegionResponse | None = None,
        dev_by_account: ListingDevByAccountResponse | None = None,
    ) -> None:
        """Initialize the CloudIOTGateway."""
        self.mammotion_http: MammotionHTTP | None = None
        self._app_key = APP_KEY
        self._app_secret = APP_SECRET
        self.domain = ALIYUN_DOMAIN

        self._client_id = self.generate_hardware_string(8)  # 8 characters
        self._device_sn = self.generate_hardware_string(32)  # 32 characters
        self._utdid = self.generate_hardware_string(32)  # 32 characters
        self._connect_response = connect_response
        self._login_by_oauth_response = login_by_oauth_response
        self._aep_response = aep_response
        self._session_by_authcode_response = session_by_authcode_response
        self._region_response = region_response
        self._devices_by_account_response = dev_by_account

    @staticmethod
    def generate_random_string(length: int):
        """Generate a random string of specified length."""
        characters = string.ascii_letters + string.digits
        return "".join(random.choice(characters) for _ in range(length))

    @staticmethod
    def generate_hardware_string(length: int) -> str:
        """Generate hardware string that is consistent per device."""
        hashed_uuid = hashlib.sha1(f"{uuid.getnode()}".encode()).hexdigest()
        return "".join(itertools.islice(itertools.cycle(hashed_uuid), length))

    def sign(self, data):
        """Generate signature for the given data."""
        keys = ["appKey", "clientId", "deviceSn", "timestamp"]
        concatenated_str = ""
        for key in keys:
            concatenated_str += f"{key}{data.get(key, '')}"

        logger.debug("sign(), toSignStr = %s", concatenated_str)

        return hmac.new(
            self._app_secret.encode("utf-8"),
            concatenated_str.encode("utf-8"),
            hashlib.sha1,
        ).hexdigest()

    def get_region(self, country_code: str, auth_code: str):
        """Get the region based on country code and auth code."""
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self.domain,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.2", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "authCode": auth_code,
                "type": "THIRD_AUTHCODE",
                "countryCode": country_code,
            },
            request=request,
            version="1.0",
        )

        # send request
        response = client.do_request("/living/account/region/get", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = json.loads(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in getting regions: " + response_body_dict["msg"])

        self._region_response = RegionResponse.from_dict(response_body_dict)
        logger.debug("Endpoint: %s", self._region_response.data.mqttEndpoint)

        return response.body

    def aep_handle(self):
        """Handle AEP authentication."""
        aep_domain = self.domain

        if self._region_response.data.apiGatewayEndpoint is not None:
            aep_domain = self._region_response.data.apiGatewayEndpoint

        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=aep_domain,
        )
        client = Client(config)

        request = CommonParams(api_ver="1.0.0", language="en-US")
        logger.debug("client id %s", self._client_id)
        time_now = time.time()
        data_to_sign = {
            "appKey": self._app_key,
            "clientId": self._client_id,  # needs to be unique to device
            "deviceSn": self._device_sn,  # same here
            "timestamp": str(time_now),
        }

        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "authInfo": {
                    "clientId": self._client_id,
                    "sign": self.sign(data_to_sign),
                    "deviceSn": self._device_sn,
                    "timestamp": str(time_now),
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        response = client.do_request("/app/aepauth/handle", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        response_body_str = response.body.decode("utf-8")

        response_body_dict = json.loads(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in getting mqtt credentials: " + response_body_dict["msg"])

        self._aep_response = AepResponse.from_dict(response_body_dict)

        logger.debug(response_body_dict)

        return response.body

    async def connect(self):
        """Connect to the Aliyun Cloud IoT Gateway."""
        region_url = "sdk.openaccount.aliyun.com"
        time_now = time.time()
        async with ClientSession() as session:
            headers = {
                "host": region_url,
                "date": UtilClient.get_date_utcstring(),
                "x-ca-nonce": UtilClient.get_nonce(),
                "x-ca-key": self._app_key,
                "x-ca-signaturemethod": "HmacSHA256",
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
                "user-agent": UtilClient.get_user_agent(None),
            }

            _bodyParam = {
                "context": {
                    "sdkVersion": "3.4.2",
                    "platformName": "android",
                    "netType": "wifi",
                    "appKey": self._app_key,
                    "yunOSId": "",
                    "appVersion": APP_VERSION,
                    "utDid": self._utdid,
                    "appAuthToken": self._utdid,  # ???
                    "securityToken": self._utdid,  # ???
                },
                "config": {"version": 0, "lastModify": 0},
                "device": {
                    "model": "sdk_gphone_x86_arm",
                    "brand": "goldfish_x86",
                    "platformVersion": "30",
                },
            }

            # Get sign header
            dic = headers.copy()
            for key in MOVE_HEADERS:
                dic.pop(key, None)

            keys = sorted(dic.keys())
            sign_headers = ",".join(keys)
            header = "".join(f"{k}:{dic[k]}\n" for k in keys).strip()

            headers["x-ca-signature-headers"] = sign_headers
            string_to_sign = "POST\n{}\n\n{}\n{}\n{}\n/api/prd/connect.json?request={}".format(
                headers["accept"],
                headers["content-type"],
                headers["date"],
                header,
                json.dumps(_bodyParam, separators=(",", ":")),
            )

            hash_val = hmac.new(
                self._app_secret.encode("utf-8"),
                string_to_sign.encode("utf-8"),
                hashlib.sha256,
            ).digest()
            signature = base64.b64encode(hash_val).decode("utf-8")
            headers["x-ca-signature"] = signature

            async with session.post(
                f"https://{region_url}/api/prd/connect.json",
                headers=headers,
                params={"request": json.dumps(_bodyParam, separators=(",", ":"))},
            ) as resp:
                data = await resp.json()
                logger.debug(data)
                if resp.status == 200:
                    self._connect_response = ConnectResponse.from_dict(data)
                    return self._connect_response
                raise LoginException(data)

    async def login_by_oauth(self, country_code: str, auth_code: str):
        """Login by OAuth."""
        region_url = self._region_response.data.oaApiGatewayEndpoint

        async with ClientSession() as session:
            headers = {
                "host": region_url,
                "date": UtilClient.get_date_utcstring(),
                "x-ca-nonce": UtilClient.get_nonce(),
                "x-ca-key": self._app_key,
                "x-ca-signaturemethod": "HmacSHA256",
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded; charset=utf-8",
                "user-agent": UtilClient.get_user_agent(None),
                "vid": self._connect_response.data.vid,
            }

            _bodyParam = {
                "country": country_code,
                "authCode": auth_code,
                "oauthPlateform": 23,
                "oauthAppKey": self._app_key,
                "riskControlInfo": {
                    "appID": "com.agilexrobotics",
                    "appAuthToken": "",
                    "signType": "RSA",
                    "sdkVersion": "3.4.2",
                    "utdid": self._utdid,
                    "umidToken": self._utdid,
                    "deviceId": self._connect_response.data.data.device.data.deviceId,
                    "USE_OA_PWD_ENCRYPT": "true",
                    "USE_H5_NC": "true",
                },
            }

            # Get sign header
            dic = headers.copy()
            for key in MOVE_HEADERS:
                dic.pop(key, None)

            keys = sorted(dic.keys())
            sign_headers = ",".join(keys)
            header = "".join(f"{k}:{dic[k]}\n" for k in keys).strip()

            headers["x-ca-signature-headers"] = sign_headers
            string_to_sign = "POST\n{}\n\n{}\n{}\n{}\n/api/prd/loginbyoauth.json?{}".format(
                headers["accept"],
                headers["content-type"],
                headers["date"],
                header,
                f"loginByOauthRequest={json.dumps(_bodyParam, separators=(",", ":"))}",
            )

            hash_val = hmac.new(
                self._app_secret.encode("utf-8"),
                string_to_sign.encode("utf-8"),
                hashlib.sha256,
            ).digest()
            signature = base64.b64encode(hash_val).decode("utf-8")
            headers["x-ca-signature"] = signature
            async with session.post(
                f"https://{region_url}/api/prd/loginbyoauth.json",
                headers=headers,
                data={"loginByOauthRequest": json.dumps(_bodyParam, separators=(",", ":"))},
            ) as resp:
                data = await resp.json()
                logger.debug(data)
                if resp.status == 200:
                    self._login_by_oauth_response = LoginByOAuthResponse.from_dict(data)
                    return self._login_by_oauth_response
                raise LoginException(data)

    def session_by_auth_code(self):
        """Create a session by auth code."""
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.4", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "request": {
                    "authCode": self._login_by_oauth_response.data.data.loginSuccessResult.sid,
                    "accountType": "OA_SESSION",
                    "appKey": self._app_key,
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        response = client.do_request(
            "/account/createSessionByAuthCode",
            "https",
            "POST",
            None,
            body,
            RuntimeOptions(),
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = json.loads(response_body_str)

        session_by_auth = SessionByAuthCodeResponse.from_dict(response_body_dict)

        if int(session_by_auth.code) != 200:
            raise Exception("Error in creating session: " + response_body_str)

        if session_by_auth.data.identityId is None:
            raise Exception("Error in creating session: " + response_body_str)

        self._session_by_authcode_response = session_by_auth
        self._iot_token_issued_at = int(time.time())

        return response.body

    def sign_out(self) -> None:
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.4", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "request": {
                    "refreshToken": self._session_by_authcode_response.data.refreshToken,
                    "identityId": self._session_by_authcode_response.data.identityId,
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        # possibly need to do this ourselves
        response = client.do_request(
            "/iotx/account/invalidSession",
            "https",
            "POST",
            None,
            body,
            RuntimeOptions(),
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = json.loads(response_body_str)
        logger.debug(response_body_dict)
        return response_body_dict

    def check_or_refresh_session(self):
        """Check or refresh the session."""
        logger.debug("Trying to refresh token")
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )
        client = Client(config)

        # build request
        request = CommonParams(api_ver="1.0.4", language="en-US")
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={
                "request": {
                    "refreshToken": self._session_by_authcode_response.data.refreshToken,
                    "identityId": self._session_by_authcode_response.data.identityId,
                }
            },
            request=request,
            version="1.0",
        )

        # send request
        # possibly need to do this ourselves
        response = client.do_request(
            "/account/checkOrRefreshSession",
            "https",
            "POST",
            None,
            body,
            RuntimeOptions(),
        )
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = json.loads(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            logger.error(response_body_dict)
            self.sign_out()
            raise CheckSessionException("Error check or refresh token: " + response_body_dict.__str__())

        session = SessionByAuthCodeResponse.from_dict(response_body_dict)
        session_data = session.data

        if (
            session_data.identityId is None
            or session_data.refreshTokenExpire is None
            or session_data.iotToken is None
            or session_data.iotTokenExpire is None
            or session_data.refreshToken is None
        ):
            raise Exception("Error check or refresh token: Parameters not correct")

        self._session_by_authcode_response = session
        self._iot_token_issued_at = int(time.time())

    def list_binding_by_account(self) -> ListingDevByAccountResponse:
        """List bindings by account."""
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.8",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={"pageSize": 100, "pageNo": 1},
            request=request,
            version="1.0",
        )

        # send request
        response = client.do_request("/uc/listBindingByAccount", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = json.loads(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in creating session: " + response_body_dict["msg"])

        self._devices_by_account_response = ListingDevByAccountResponse.from_dict(response_body_dict)
        return self._devices_by_account_response

    def list_binding_by_dev(self, iot_id: str):
        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.8",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )
        body = IoTApiRequest(
            id=str(uuid.uuid4()),
            params={"pageSize": 100, "pageNo": 1, "iotId": iot_id},
            request=request,
            version="1.0",
        )

        # send request
        response = client.do_request("/uc/listBindingByDev", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)

        # Decode the response body
        response_body_str = response.body.decode("utf-8")

        # Load the JSON string into a dictionary
        response_body_dict = json.loads(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            raise Exception("Error in creating session: " + response_body_dict["msg"])

        self._devices_by_account_response = ListingDevByAccountResponse.from_dict(response_body_dict)
        return self._devices_by_account_response

    def send_cloud_command(self, iot_id: str, command: bytes) -> str:
        """Send a cloud command to the specified IoT device."""

        if command is None:
            raise Exception("Command is missing / None")

        """Check if iotToken is expired"""
        if self._iot_token_issued_at + self._session_by_authcode_response.data.iotTokenExpire <= (
            int(time.time()) + (5 * 3600)
        ):
            """Token expired - Try to refresh - Check if refreshToken is not expired"""
            if self._iot_token_issued_at + self._session_by_authcode_response.data.refreshTokenExpire > (
                int(time.time())
            ):
                self.check_or_refresh_session()
            else:
                raise AuthRefreshException("Refresh token expired. Please re-login")

        config = Config(
            app_key=self._app_key,
            app_secret=self._app_secret,
            domain=self._region_response.data.apiGatewayEndpoint,
        )

        client = Client(config)

        # build request
        request = CommonParams(
            api_ver="1.0.5",
            language="en-US",
            iot_token=self._session_by_authcode_response.data.iotToken,
        )

        # TODO move to using InvokeThingServiceRequest()

        message_id = str(uuid.uuid4())

        body = IoTApiRequest(
            id=message_id,
            params={
                "args": {"content": self.converter.printBase64Binary(command)},
                "identifier": "device_protobuf_sync_service",
                "iotId": f"{iot_id}",
            },
            request=request,
            version="1.0",
        )
        logger.debug(self.converter.printBase64Binary(command))
        # send request
        response = client.do_request("/thing/service/invoke", "https", "POST", None, body, RuntimeOptions())
        logger.debug(response.status_message)
        logger.debug(response.headers)
        logger.debug(response.status_code)
        logger.debug(response.body)
        logger.debug(iot_id)

        response_body_str = response.body.decode("utf-8")
        response_body_dict = json.loads(response_body_str)

        if int(response_body_dict.get("code")) != 200:
            logger.error(
                "Error in sending cloud command: %s - %s",
                str(response_body_dict.get("code")),
                str(response_body_dict.get("message")),
            )
            if response_body_dict.get("code") == 20056:
                logger.debug("Gateway timeout.")
                raise GatewayTimeoutException(response_body_dict.get("code"))

            if response_body_dict.get("code") == 29003:
                logger.debug(self._session_by_authcode_response.data.identityId)
                self.sign_out()
                raise SetupException(response_body_dict.get("code"))
            if response_body_dict.get("code") == 6205:
                raise DeviceOfflineException(response_body_dict.get("code"))

        return message_id

    @property
    def devices_by_account_response(self):
        return self._devices_by_account_response

    def set_http(self, mammotion_http: MammotionHTTP) -> None:
        self.mammotion_http = mammotion_http

    @property
    def region_response(self) -> RegionResponse:
        return self._region_response

    @property
    def aep_response(self) -> AepResponse:
        return self._aep_response

    @property
    def session_by_authcode_response(self) -> SessionByAuthCodeResponse:
        return self._session_by_authcode_response

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def login_by_oauth_response(self) -> LoginByOAuthResponse:
        return self._login_by_oauth_response

    @property
    def connect_response(self) -> ConnectResponse:
        return self._connect_response



# === sendOrderMsg_Driver ===
from abc import ABC
from logging import getLogger
import time

from pymammotion.mammotion.commands.abstract_message import AbstractMessage
from pymammotion.proto import (
    DrvKnifeHeight,
    DrvMotionCtrl,
    DrvMowCtrlByHand,
    DrvSrSpeed,
    LubaMsg,
    MctlDriver,
    MsgAttr,
    MsgCmdType,
    MsgDevice,
    RtkCfgReqT,
    RtkSysMaskQueryT,
)

logger = getLogger(__name__)


class MessageDriver(AbstractMessage, ABC):
    def send_order_msg_driver(self, driver) -> bytes:
        return LubaMsg(
            msgtype=MsgCmdType.EMBED_DRIVER,
            sender=MsgDevice.DEV_MOBILEAPP,
            rcver=self.get_msg_device(MsgCmdType.EMBED_DRIVER, MsgDevice.DEV_MAINCTL),
            msgattr=MsgAttr.REQ,
            timestamp=round(time.time() * 1000),
            seqs=1,
            version=1,
            subtype=1,
            driver=driver,
        ).SerializeToString()

    def set_blade_height(self, height: int):
        logger.debug(f"Send knife height height={height}")
        build = MctlDriver(todev_knife_height_set=DrvKnifeHeight(knife_height=height))
        logger.debug(f"Send command--Knife motor height setting height={height}")
        return self.send_order_msg_driver(build)

    def set_speed(self, speed: float):
        logger.debug(f"{self.get_device_name()} set speed, {speed}")
        build = MctlDriver(bidire_speed_read_set=DrvSrSpeed(speed=speed, rw=1))
        logger.debug(f"Send command--Speed setting speed={speed}")
        return self.send_order_msg_driver(build)

    def syn_nav_star_point_data(self, sat_system: int):
        build = MctlDriver(rtk_sys_mask_query=RtkSysMaskQueryT(sat_system=sat_system))
        logger.debug(f"Send command--Navigation satellite frequency point synchronization={sat_system}")
        return self.send_order_msg_driver(build)

    def set_nav_star_point(self, cmd_req: str):
        build = MctlDriver(rtk_cfg_req=RtkCfgReqT(cmd_req=cmd_req, cmd_length=len(cmd_req) - 1))
        logger.debug(f"Send command--Navigation satellite frequency point setting={cmd_req}")
        logger.debug(
            f"Navigation satellite setting, Send command--Navigation satellite frequency point setting={cmd_req}"
        )
        return self.send_order_msg_driver(build)

    def get_speed(self):
        build = MctlDriver(bidire_speed_read_set=DrvSrSpeed(rw=0))
        logger.debug("Send command--Get speed value")
        return self.send_order_msg_driver(build)

    def operate_on_device(
        self,
        main_ctrl: int,
        cut_knife_ctrl: int,
        cut_knife_height: int,
        max_run_speed: float,
    ):
        build = MctlDriver(
            mow_ctrl_by_hand=DrvMowCtrlByHand(
                main_ctrl=main_ctrl,
                cut_knife_ctrl=cut_knife_ctrl,
                cut_knife_height=cut_knife_height,
                max_run_speed=max_run_speed,
            )
        )
        logger.debug(
            f"Send command--Manual mowing command, main_ctrl:{main_ctrl}, cut_knife_ctrl:{cut_knife_ctrl}, "
            f"cut_knife_height:{cut_knife_height}, max_run_speed:{max_run_speed}"
        )

        return self.send_order_msg_driver(build)

    def send_movement(self, linear_speed: int, angular_speed: int) -> bytes:
        logger.debug(f"Control command print, linearSpeed={
        linear_speed} // angularSpeed={angular_speed}")
        return self.send_order_msg_driver(
            MctlDriver(
                todev_devmotion_ctrl=DrvMotionCtrl(set_linear_speed=linear_speed, set_angular_speed=angular_speed)
            )
        )



from dataclasses import dataclass, field


@dataclass
class RangeLimit:
    min: float
    max: float


@dataclass
class DeviceLimits:
    blade_height: RangeLimit = field(default_factory=RangeLimit)
    working_speed: RangeLimit = field(default_factory=RangeLimit)
    working_path: RangeLimit = field(default_factory=RangeLimit)
    work_area_num_max: int = 60
    display_image_type: int = 0

    def to_dict(self) -> dict:
        """Convert the device limits to a dictionary format."""
        return {
            "blade_height": {"min": self.blade_height.min, "max": self.blade_height.max},
            "working_speed": {"min": self.working_speed.min, "max": self.working_speed.max},
            "working_path": {"min": self.working_path.min, "max": self.working_path.max},
            "work_area_num_max": self.work_area_num_max,
            "display_image_type": self.display_image_type,
        }

    @classmethod
    def from_dict(cls, data: dict) -> "DeviceLimits":
        """Create a DeviceLimits instance from a dictionary."""
        return cls(
            blade_height=RangeLimit(min=data["blade_height"]["min"], max=data["blade_height"]["max"]),
            working_speed=RangeLimit(min=data["working_speed"]["min"], max=data["working_speed"]["max"]),
            working_path=RangeLimit(min=data["working_path"]["min"], max=data["working_path"]["max"]),
            work_area_num_max=data["work_area_num_max"],
            display_image_type=data["display_image_type"],
        )

    def validate(self) -> bool:
        """Validate that all ranges are logical (min <= max)."""
        return all(
            [
                self.blade_height.min <= self.blade_height.max,
                self.working_speed.min <= self.working_speed.max,
                self.working_path.min <= self.working_path.max,
                self.work_area_num_max > 0,
                self.display_image_type in (0, 1),
            ]
        )



from enum import IntEnum

from pymammotion.data.model.report_info import ConnectData


class bleOrderCmd:
    allpowerfullRW = 67
    alongBorder = 9
    areaAndTimeAndPathUpdate = 27
    autoUnderPile = 65
    batteryValueUpdate = 5
    bleAlive = 59
    cancelCurrentRecord = 62
    cancelLogUpdate = 45
    cancelPauseExecuteTask = 47
    checkFramwareVersion = 41
    checkFramwareVersionCallBack = 42
    checkSoftVersion = 35
    checkSoftVersionCallBack = 36
    closeKinfe = 50
    close_clear_connect_current_wifi = 77
    deletingConnectionPath = 22
    deviceOrderResponse = 46
    deviceStatusUpdate = 31
    endChannelLine = 15
    endDrawBarrier = 13
    endDrawBorder = 11
    errorCodeUpdate = 34
    framwarePackageCallBack = 40
    framwareSuccessUpdate = 48
    framwareUpdateCallBack = 38
    generateRouteInformation = 210
    getAreaData = 209
    getDeiveBorderState = 64
    getDeviceInfo = 63
    getDeviceLogInfo = 43
    getHashList = 208
    getrecordwifi = 69
    initPointUpdate = 3
    job_plan_setting_read_delete = 78
    job_plan_setting_read_delete_unable_time = 79
    kinfeState = 51
    knifeHightUpdate = 29
    logProgressUpdate = 44
    openKinfe = 49
    optimizationBorderUpadate = 17
    optimizationChannalLineUpdate = 19
    optimizationObstacleUpdate = 18
    originLagLog = 52
    pauseExecuteTask = 7
    removeObstaclesOrObstructions = 21
    replyOptimizationPackage = 20
    resetBaseStation = 61
    responseDevice = 52
    retrunGenerateRouteInformation = 211
    returnCharge = 8
    rtloactionUpdate = 4
    saveTask = 16
    sendBorderPackage = 21
    sendBorderPackageCallBack = 22
    sendChannalLinePackage = 25
    sendChannalLinePackageCallBack = 26
    sendContrlCallBack = 2
    sendControl = 1
    sendExecuteTask = 6
    sendFramwarePackage = 39
    sendFramwareUpdate = 37
    sendObstaclePackage = 23
    sendObstaclePackageCallBack = 24
    sendPlan = 32
    setKnifeHight = 28
    setMaxSpeed = 33
    startChannelLine = 14
    startDrawBarrier = 12
    startDrawBorder = 10
    startWorkOrder = 60
    startjob = 212
    synTime = 57
    task = 213
    taskProgressUpdate = 30
    testPW = 52
    wificonnectinfoupdate = 68
    wirteReadSpeed = 66


class SystemUpdateBuf:
    BATTERY_STATE_INDEX = 2
    CHARGE_POS_VAILD_INDEX = 9
    CHARGE_POS_X_F_INDEX = 7
    CHARGE_POS_Y_F_INDEX = 8
    CHARGE_TOWARD_INDEX = 3
    ERR_CODE_10_INDEX = 21
    ERR_CODE_1_INDEX = 3
    ERR_CODE_2_INDEX = 5
    ERR_CODE_3_INDEX = 7
    ERR_CODE_4_INDEX = 9
    ERR_CODE_5_INDEX = 11
    ERR_CODE_6_INDEX = 13
    ERR_CODE_7_INDEX = 15
    ERR_CODE_8_INDEX = 17
    ERR_CODE_9_INDEX = 19
    ERR_CODE_CNT_INDEX = 2
    ERR_CODE_ID_INDEX = 0
    ERR_CODE_LEN_INDEX = 1
    ERR_CODE_STAMP_10_INDEX = 22
    ERR_CODE_STAMP_1_INDEX = 4
    ERR_CODE_STAMP_2_INDEX = 6
    ERR_CODE_STAMP_3_INDEX = 8
    ERR_CODE_STAMP_4_INDEX = 10
    ERR_CODE_STAMP_5_INDEX = 12
    ERR_CODE_STAMP_6_INDEX = 14
    ERR_CODE_STAMP_7_INDEX = 16
    ERR_CODE_STAMP_8_INDEX = 18
    ERR_CODE_STAMP_9_INDEX = 20
    SU_LAT_D_INDEX = 5
    SU_LON_D_INDEX = 6
    SU_SPEED_F_INDEX = 4
    SYSTEM_ERR_CODE_INDEX_END = 23
    SYSTEM_INIT_CONFIG_ID_INDEX = 0
    SYSTEM_INIT_CONFIG_INDEX_END = 10
    SYSTEM_INIT_CONFIG_LEN_INDEX = 1
    SYSTEM_ZONE_STATE_INDEX_END = 22
    ZONE_STATE_1_INDEX = 2
    ZONE_STATE_ID_INDEX = 0
    ZONE_STATE_LEN_INDEX = 1


class SystemRapidStateTunnelIndex(IntEnum):
    DIS_CAR_RTK_STARS_INDEX = 15
    DIS_RTK_STATUS_INDEX = 13
    L1_SATS_INDEX = 2
    L2_SATS_INDEX = 6
    POS_LEVEL_INDEX = 1
    POS_TYPE_INDEX = 10
    RAPID_WORK_STATE_VER_INDEX = 12
    REAL_POS_X_F_INDEX = 7
    REAL_POS_Y_F_INDEX = 8
    REAL_TOWARD_F_INDEX = 9
    RTK_AGE_F_INDEX = 3
    SIGNAL_QUALITY_INDEX = 0
    TOP4_TOTAL_MEAN_INDEX = 14
    VEL_MEAN_F_INDEX = 5
    VEL_TOP_F_INDEX = 4
    VSIAM_STATE_INDEX = 16
    ZONE_HASH_INDEX = 11


class SystemTardStateTunnel:
    APP_CONNECTED_INFO = 27
    BATTERY_VAL_INDEX = 2
    BOL_HASH_INDEX = 6
    BREAK_POINT_HASH_INDEX = 11
    BREAK_POINT_INFO_INDEX = 10
    BREAK_POINT_X_F_INDEX = 12
    BREAK_POINT_Y_F_INDEX = 13
    CHARGE_STATE_INDEX = 1
    CUT_HEIGHT_INDEX = 3
    DEVICE_OLD_STATUS_INDEX = 25
    DEVICE_STATE_INDEX = 0
    DRAWING_RTK_BAD_OLD_STATE_INDEX = 29
    MAINTAIN_TOTAL_BATTERY_CYCLES_INDEX = 32
    MAINTAIN_TOTAL_MILEAGE_INDEX = 30
    MAINTAIN_TOTAL_MOWING_TIME_INDEX = 31
    MOW_RUN_SPEED_INDEX = 24
    PATH_HASH_INDEX = 7
    PATH_POS_X_F_INDEX = 15
    PATH_POS_Y_F_INDEX = 16
    PLAN_STATE_INDEX = 5
    REAL_PATH_INDEX = 14
    RTK_LORA_NUM_CHANNEL = 35
    RTK_LORA_NUM_LOC_ID = 36
    RTK_LORA_NUM_NET_ID = 37
    RTK_LORA_NUM_SCAN = 34
    RTK_RESTARTING_INDEX = 28
    RTK_STARS_NUM = 33
    RTK_STATUS = 38
    SENSOR_STATE_INDEX = 4
    SYSTEM_TIME_STAMP = 26
    TARD_WORK_STATE_END = 21
    TARD_WORK_STATE_VER_INDEX = 22
    TASK_AREA_INDEX = 9
    TASK_PROGRESS_INDEX = 8
    TEST_SWITCH_STATE_INDEX = 23
    UB_ERR_CODE_HASH_INDEX = 20
    UB_INIT_CONFIG_HASH_INDEX = 19
    UB_REAL_PATH_HASH_INDEX = 18
    UB_ZONE_STATE_HASH_INDEX = 17


class WorkMode:
    MODE_NOT_ACTIVE = 0
    MODE_ONLINE = 1
    MODE_OFFLINE = 2
    MODE_DISABLE = 8
    MODE_INITIALIZATION = 10
    MODE_READY = 11
    MODE_WORKING = 13
    MODE_RETURNING = 14
    MODE_CHARGING = 15
    MODE_UPDATING = 16
    MODE_LOCK = 17
    MODE_PAUSE = 19
    MODE_MANUAL_MOWING = 20
    MODE_UPDATE_SUCCESS = 22
    MODE_OTA_UPGRADE_FAIL = 23
    MODE_JOB_DRAW = 31
    MODE_OBSTACLE_DRAW = 32
    MODE_CHANNEL_DRAW = 34
    MODE_ERASER_DRAW = 35
    MODE_EDIT_BOUNDARY = 36
    MODE_LOCATION_ERROR = 37
    MODE_BOUNDARY_JUMP = 38
    MODE_CHARGING_PAUSE = 39


def device_connection(connect: ConnectData) -> str:
    """Return string representation of device connection."""

    if connect.wifi_rssi != 0 and connect.ble_rssi != 0:
        return "WIFI/BLE"

    if connect.connect_type == 2 or connect.used_net == "NET_USED_TYPE_WIFI" or connect.wifi_rssi != 0:
        return "WIFI"

    if connect.connect_type == 1 or connect.used_net == "NET_USED_TYPE_MNET":
        return "3G/4G"

    if connect.ble_rssi != 0:
        return "BLE"

    return "None"


def device_mode(value: int) -> str:
    """Return the mode corresponding to the given value.

    This function takes a value and returns the corresponding mode from a
    predefined dictionary.

    Args:
        value (int): The value for which mode needs to be determined.

    Returns:
        str: The mode corresponding to the input value. Returns "Invalid mode" if no
            mode is found.

    """

    modes = {
        0: "MODE_NOT_ACTIVE",
        1: "MODE_ONLINE",
        2: "MODE_OFFLINE",
        8: "MODE_DISABLE",
        10: "MODE_INITIALIZATION",
        11: "MODE_READY",
        12: "MODE_UNCONNECTED",
        13: "MODE_WORKING",
        14: "MODE_RETURNING",
        15: "MODE_CHARGING",
        16: "MODE_UPDATING",
        17: "MODE_LOCK",
        19: "MODE_PAUSE",
        20: "MODE_MANUAL_MOWING",
        22: "MODE_UPDATE_SUCCESS",
        23: "MODE_OTA_UPGRADE_FAIL",
        31: "MODE_JOB_DRAW",
        32: "MODE_OBSTACLE_DRAW",
        34: "MODE_CHANNEL_DRAW",
        35: "MODE_ERASER_DRAW",
        36: "MODE_EDIT_BOUNDARY",
        37: "MODE_LOCATION_ERROR",
        38: "MODE_BOUNDARY_JUMP",
        39: "MODE_CHARGING_PAUSE",
    }
    return modes.get(value, "Invalid mode")


class PosType(IntEnum):
    """Position of the robot."""

    AREA_BORDER_ON = 7
    AREA_INSIDE = 1
    AREA_OUT = 0
    CHANNEL_AREA_OVERLAP = 9
    CHANNEL_ON = 3
    CHARGE_ON = 5
    DUMPING_AREA_INSIDE = 8
    DUMPING_OUTSIDE = 10
    OBS_ON = 2
    TURN_AREA_INSIDE = 4
    VIRTUAL_INSIDE = 6


def camera_brightness(value: int) -> str:
    """Return the brightness corresponding to the given value."""
    modes = {
        0: "Dark",
        1: "Light",
    }
    return modes.get(value, "Invalid mode")



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

ACTIVATION_FAILED: MUL_VIDEO_ERROR_CODE
ALL: MUL_CAMERA_POSITION
CELLULAR_RESTRICTION: MUL_VIDEO_ERROR_CODE
CREATE_CHANNEL_FAILED: MUL_VIDEO_ERROR_CODE
DESCRIPTOR: _descriptor.FileDescriptor
ENGLISH: MUL_LANGUAGE
GERMAN: MUL_LANGUAGE
HW_ERROR: MUL_WIPER_ERROR_CODE
LEFT: MUL_CAMERA_POSITION
NAVIGATION_WORK_FORBID: MUL_WIPER_ERROR_CODE
NETWORK_NOT_AVAILABLE: MUL_VIDEO_ERROR_CODE
PARAM_INVAILD: MUL_VIDEO_ERROR_CODE
REAR: MUL_CAMERA_POSITION
RIGHT: MUL_CAMERA_POSITION
SET_SUCCESS: MUL_WIPER_ERROR_CODE
SUCCESS: MUL_VIDEO_ERROR_CODE
UNRECOGNIZED: MUL_LANGUAGE

class MulAudioCfg(_message.Message):
    __slots__ = ["au_language", "au_switch"]
    AU_LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    AU_SWITCH_FIELD_NUMBER: _ClassVar[int]
    au_language: MUL_LANGUAGE
    au_switch: int
    def __init__(self, au_switch: _Optional[int] = ..., au_language: _Optional[_Union[MUL_LANGUAGE, str]] = ...) -> None: ...

class MulSetAudio(_message.Message):
    __slots__ = ["at_switch", "au_language"]
    AT_SWITCH_FIELD_NUMBER: _ClassVar[int]
    AU_LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    at_switch: int
    au_language: MUL_LANGUAGE
    def __init__(self, at_switch: _Optional[int] = ..., au_language: _Optional[_Union[MUL_LANGUAGE, str]] = ...) -> None: ...

class MulSetVideo(_message.Message):
    __slots__ = ["position", "vi_switch"]
    POSITION_FIELD_NUMBER: _ClassVar[int]
    VI_SWITCH_FIELD_NUMBER: _ClassVar[int]
    position: MUL_CAMERA_POSITION
    vi_switch: int
    def __init__(self, position: _Optional[_Union[MUL_CAMERA_POSITION, str]] = ..., vi_switch: _Optional[int] = ...) -> None: ...

class MulSetVideoAck(_message.Message):
    __slots__ = ["error_code"]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    error_code: MUL_VIDEO_ERROR_CODE
    def __init__(self, error_code: _Optional[_Union[MUL_VIDEO_ERROR_CODE, str]] = ...) -> None: ...

class MulSetWiper(_message.Message):
    __slots__ = ["round"]
    ROUND_FIELD_NUMBER: _ClassVar[int]
    round: int
    def __init__(self, round: _Optional[int] = ...) -> None: ...

class MulSetWiperAck(_message.Message):
    __slots__ = ["error_code"]
    ERROR_CODE_FIELD_NUMBER: _ClassVar[int]
    error_code: MUL_WIPER_ERROR_CODE
    def __init__(self, error_code: _Optional[_Union[MUL_WIPER_ERROR_CODE, str]] = ...) -> None: ...

class SocMul(_message.Message):
    __slots__ = ["audio_cfg", "set_audio", "set_video", "set_video_ack", "set_wiper", "set_wiper_ack"]
    AUDIO_CFG_FIELD_NUMBER: _ClassVar[int]
    SET_AUDIO_FIELD_NUMBER: _ClassVar[int]
    SET_VIDEO_ACK_FIELD_NUMBER: _ClassVar[int]
    SET_VIDEO_FIELD_NUMBER: _ClassVar[int]
    SET_WIPER_ACK_FIELD_NUMBER: _ClassVar[int]
    SET_WIPER_FIELD_NUMBER: _ClassVar[int]
    audio_cfg: MulAudioCfg
    set_audio: MulSetAudio
    set_video: MulSetVideo
    set_video_ack: MulSetVideoAck
    set_wiper: MulSetWiper
    set_wiper_ack: MulSetWiperAck
    def __init__(self, set_audio: _Optional[_Union[MulSetAudio, _Mapping]] = ..., audio_cfg: _Optional[_Union[MulAudioCfg, _Mapping]] = ..., set_video: _Optional[_Union[MulSetVideo, _Mapping]] = ..., set_video_ack: _Optional[_Union[MulSetVideoAck, _Mapping]] = ..., set_wiper: _Optional[_Union[MulSetWiper, _Mapping]] = ..., set_wiper_ack: _Optional[_Union[MulSetWiperAck, _Mapping]] = ...) -> None: ...

class MUL_LANGUAGE(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_CAMERA_POSITION(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_VIDEO_ERROR_CODE(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class MUL_WIPER_ERROR_CODE(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import asyncio
from asyncio import TimerHandle
import base64
from collections import deque
from collections.abc import Awaitable, Callable
import json
import logging
from typing import Any, cast

import betterproto

from pymammotion import CloudIOTGateway, MammotionMQTT
from pymammotion.aliyun.cloud_gateway import DeviceOfflineException
from pymammotion.aliyun.model.dev_by_account_response import Device
from pymammotion.data.mqtt.event import ThingEventMessage
from pymammotion.data.mqtt.properties import ThingPropertiesMessage
from pymammotion.data.mqtt.status import ThingStatusMessage
from pymammotion.data.state_manager import StateManager
from pymammotion.event.event import DataEvent
from pymammotion.mammotion.commands.mammotion_command import MammotionCommand
from pymammotion.mammotion.devices.base import MammotionBaseDevice
from pymammotion.mqtt.mammotion_future import MammotionFuture
from pymammotion.proto import LubaMsg, has_field

_LOGGER = logging.getLogger(__name__)


class MammotionCloud:
    """Per account MQTT cloud."""

    def __init__(self, mqtt_client: MammotionMQTT, cloud_client: CloudIOTGateway) -> None:
        self.cloud_client = cloud_client
        self.loop = asyncio.get_event_loop()
        self.is_ready = False
        self.command_queue = asyncio.Queue()
        self._waiting_queue = deque()
        self.mqtt_message_event = DataEvent()
        self.mqtt_properties_event = DataEvent()
        self.mqtt_status_event = DataEvent()
        self.on_ready_event = DataEvent()
        self.on_disconnected_event = DataEvent()
        self.on_connected_event = DataEvent()
        self._operation_lock = asyncio.Lock()
        self._mqtt_client = mqtt_client
        self._mqtt_client.on_connected = self.on_connected
        self._mqtt_client.on_disconnected = self.on_disconnected
        self._mqtt_client.on_message = self._on_mqtt_message
        self._mqtt_client.on_ready = self.on_ready

        # temporary for testing only
        # self._start_sync_task = self.loop.call_later(30, lambda: asyncio.ensure_future(self.start_sync(0)))

    async def on_ready(self) -> None:
        loop = asyncio.get_event_loop()
        loop.create_task(self.process_queue())
        await self.on_ready_event.data_event(None)

    def is_connected(self) -> bool:
        return self._mqtt_client.is_connected

    def disconnect(self) -> None:
        self._mqtt_client.disconnect()

    def connect_async(self) -> None:
        self._mqtt_client.connect_async()

    def send_command(self, iot_id: str, command: bytes) -> None:
        self._mqtt_client.get_cloud_client().send_cloud_command(iot_id, command)

    async def on_connected(self) -> None:
        """Callback for when MQTT connects."""
        await self.on_connected_event.data_event(None)

    async def on_disconnected(self) -> None:
        """Callback for when MQTT disconnects."""
        await self.on_disconnected_event.data_event(None)

    async def process_queue(self) -> None:
        while True:
            # Get the next item from the queue
            iot_id, key, command, future = await self.command_queue.get()
            try:
                # Process the command using _execute_command_locked
                result = await self._execute_command_locked(iot_id, key, command)
                # Set the result on the future
                future.set_result(result)
            except Exception as ex:
                # Set the exception on the future if something goes wrong
                future.set_exception(ex)
            finally:
                # Mark the task as done
                self.command_queue.task_done()

    async def _execute_command_locked(self, iot_id: str, key: str, command: bytes) -> bytes:
        """Execute command and read response."""
        assert self._mqtt_client is not None
        self._key = key
        _LOGGER.debug("Sending command: %s", key)

        await self.loop.run_in_executor(None, self._mqtt_client.get_cloud_client().send_cloud_command, iot_id, command)
        future = MammotionFuture(iot_id)
        self._waiting_queue.append(future)
        timeout = 5
        try:
            notify_msg = await future.async_get(timeout)
        except asyncio.TimeoutError:
            _LOGGER.debug("command_locked TimeoutError")
            notify_msg = b""

        _LOGGER.debug("%s: Message received", iot_id)

        return notify_msg

    async def _on_mqtt_message(self, topic: str, payload: str, iot_id: str) -> None:
        """Handle incoming MQTT messages."""
        _LOGGER.debug("MQTT message received on topic %s: %s, iot_id: %s", topic, payload, iot_id)

        json_str = json.dumps(payload)
        payload = json.loads(json_str)

        await self._parse_mqtt_response(topic, payload)

    async def _parse_mqtt_response(self, topic: str, payload: dict) -> None:
        """Parse the MQTT response."""
        if topic.endswith("/app/down/thing/events"):
            _LOGGER.debug("Thing event received")
            event = ThingEventMessage.from_dicts(payload)
            params = event.params
            if isinstance(params, dict) or params.identifier is None:
                _LOGGER.debug("Received dict params: %s", params)
                return
            if params.identifier == "device_protobuf_msg_event" and event.method == "thing.events":
                _LOGGER.debug("Protobuf event")
                # Call the callbacks for each cloudDevice
                await self.mqtt_message_event.data_event(event)
            if event.method == "thing.properties":
                await self.mqtt_properties_event.data_event(event)
                _LOGGER.debug(event)
        elif topic.endswith("/app/down/thing/status"):
            status = ThingStatusMessage.from_dict(payload)
            await self.mqtt_status_event.data_event(status)

    def _disconnect(self) -> None:
        """Disconnect the MQTT client."""
        self._mqtt_client.disconnect()

    @property
    def waiting_queue(self):
        return self._waiting_queue


class MammotionBaseCloudDevice(MammotionBaseDevice):
    """Base class for Mammotion Cloud devices."""

    def __init__(self, mqtt: MammotionCloud, cloud_device: Device, state_manager: StateManager) -> None:
        """Initialize MammotionBaseCloudDevice."""
        super().__init__(state_manager, cloud_device)
        self._ble_sync_task: TimerHandle | None = None
        self.stopped = False
        self.on_ready_callback: Callable[[], Awaitable[None]] | None = None
        self.loop = asyncio.get_event_loop()
        self._mqtt = mqtt
        self.iot_id = cloud_device.iotId
        self.device = cloud_device
        self._command_futures = {}
        self._commands: MammotionCommand = MammotionCommand(cloud_device.deviceName)
        self.currentID = ""
        self._mqtt.mqtt_message_event.add_subscribers(self._parse_message_for_device)
        self._mqtt.mqtt_properties_event.add_subscribers(self._parse_message_properties_for_device)
        self._mqtt.mqtt_status_event.add_subscribers(self._parse_message_status_for_device)
        self._mqtt.on_ready_event.add_subscribers(self.on_ready)
        self._mqtt.on_disconnected_event.add_subscribers(self.on_disconnect)
        self._mqtt.on_connected_event.add_subscribers(self.on_connect)
        self._state_manager.cloud_gethash_ack_callback = self.datahash_response
        self._state_manager.cloud_get_commondata_ack_callback = self.commdata_response
        self.set_queue_callback(self.queue_command)

        if self._mqtt.is_ready:
            self.run_periodic_sync_task()

    def __del__(self) -> None:
        self._mqtt.on_ready_event.remove_subscribers(self.on_ready)
        self._mqtt.on_disconnected_event.remove_subscribers(self.on_disconnect)
        self._mqtt.on_connected_event.remove_subscribers(self.on_connect)
        self._mqtt.mqtt_message_event.remove_subscribers(self._parse_message_for_device)
        self._state_manager.cloud_gethash_ack_callback = None
        self._state_manager.cloud_get_commondata_ack_callback = None
        if self._ble_sync_task:
            self._ble_sync_task.cancel()

    def set_notification_callback(self, func: Callable[[tuple[str, Any | None]], Awaitable[None]]) -> None:
        self._state_manager.cloud_on_notification_callback = func

    def set_queue_callback(self, func: Callable[[str, dict[str, Any]], Awaitable[bytes]]) -> None:
        self._state_manager.cloud_queue_command_callback = func

    async def on_ready(self) -> None:
        """Callback for when MQTT is subscribed to events."""
        if self.stopped:
            return
        try:
            if self.on_ready_callback:
                await self.on_ready_callback()
        except DeviceOfflineException:
            _LOGGER.debug("Device is offline")

    async def on_disconnect(self) -> None:
        if self._ble_sync_task:
            self._ble_sync_task.cancel()
        self._mqtt.disconnect()

    async def on_connect(self) -> None:
        await self._ble_sync()
        if self._ble_sync_task is None or self._ble_sync_task.cancelled():
            await self.run_periodic_sync_task()

    async def stop(self) -> None:
        """Stop all tasks and disconnect."""
        if self._ble_sync_task:
            self._ble_sync_task.cancel()
        self.stopped = True

    async def start(self) -> None:
        await self._ble_sync()
        if self._ble_sync_task is None or self._ble_sync_task.cancelled():
            await self.run_periodic_sync_task()
        self.stopped = False
        if not self.mqtt.is_connected():
            self.mqtt.connect_async()

    async def _ble_sync(self) -> None:
        command_bytes = self._commands.send_todev_ble_sync(3)
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._mqtt.send_command, self.iot_id, command_bytes)

    async def run_periodic_sync_task(self) -> None:
        """Send ble sync to robot."""
        try:
            if not self._mqtt._operation_lock.locked() or not self.stopped:
                await self._ble_sync()
        finally:
            if not self.stopped:
                self.schedule_ble_sync()

    def schedule_ble_sync(self) -> None:
        """Periodically sync to keep connection alive."""
        if self._mqtt is not None and self._mqtt.is_connected:
            self._ble_sync_task = self.loop.call_later(
                160, lambda: asyncio.ensure_future(self.run_periodic_sync_task())
            )

    async def queue_command(self, key: str, **kwargs: Any) -> bytes:
        # Create a future to hold the result
        _LOGGER.debug("Queueing command: %s", key)
        future = asyncio.Future()
        # Put the command in the queue as a tuple (key, command, future)
        command_bytes = getattr(self._commands, key)(**kwargs)
        await self._mqtt.command_queue.put((self.iot_id, key, command_bytes, future))
        # Wait for the future to be resolved
        try:
            return await future
        except asyncio.CancelledError:
            """Try again once."""
            future = asyncio.Future()
            await self._mqtt.command_queue.put((self.iot_id, key, command_bytes, future))

    def _extract_message_id(self, payload: dict) -> str:
        """Extract the message ID from the payload."""
        return payload.get("id", "")

    def _extract_encoded_message(self, payload: dict) -> str:
        """Extract the encoded message from the payload."""
        try:
            content = payload.get("data", {}).get("data", {}).get("params", {}).get("content", "")
            return str(content)
        except AttributeError:
            _LOGGER.error("Error extracting encoded message. Payload: %s", payload)
            return ""

    @staticmethod
    def dequeue_by_iot_id(queue, iot_id):
        for item in queue:
            if item.iot_id == iot_id:
                queue.remove(item)
                return item
        return None

    async def _parse_message_properties_for_device(self, event: ThingPropertiesMessage) -> None:
        if event.params.iotId != self.iot_id:
            return
        self.state_manager.properties(event)

    async def _parse_message_status_for_device(self, status: ThingStatusMessage) -> None:
        if status.params.iotId != self.iot_id:
            return
        self.state_manager.status(status)

    async def _parse_message_for_device(self, event: ThingEventMessage) -> None:
        _LOGGER.debug("_parse_message_for_device")
        params = event.params
        new_msg = LubaMsg()
        if event.params.iotId != self.iot_id:
            return
        binary_data = base64.b64decode(params.value.content)
        try:
            self._update_raw_data(binary_data)
            new_msg = LubaMsg().parse(binary_data)
        except (KeyError, ValueError, IndexError, UnicodeDecodeError):
            _LOGGER.exception("Error parsing message %s", binary_data)

        if (
            self._commands.get_device_product_key() == ""
            and self._commands.get_device_name() == event.params.deviceName
        ):
            self._commands.set_device_product_key(event.params.productKey)

        if betterproto.serialized_on_wire(new_msg.net):
            if new_msg.net.todev_ble_sync != 0 or has_field(new_msg.net.toapp_wifi_iot_status):
                return

        await self._state_manager.notification(new_msg)

        if len(self._mqtt.waiting_queue) > 0:
            fut: MammotionFuture = self.dequeue_by_iot_id(self._mqtt.waiting_queue, self.iot_id)
            if fut is None:
                return
            while fut.fut.cancelled() and len(self._mqtt.waiting_queue) > 0:
                fut = self.dequeue_by_iot_id(self._mqtt.waiting_queue, self.iot_id)
            if not fut.fut.cancelled():
                fut.resolve(cast(bytes, binary_data))

    @property
    def mqtt(self):
        return self._mqtt
